local L = AceLibrary("AceLocale-2.2"):new("Omen")
local Aura = AceLibrary("SpecialEvents-Aura-2.0")
local sprintf = string.format

OmenClassModuleCore = Omen:NewModule("ClassCore")
OmenClassModuleCore:SetModuleMixins("AceDebug-2.0", "AceEvent-2.0")

function OmenClassModuleCore:OnInitialize()	
	local rc, c = UnitClass("player")
	local playerClassModule = "Class-" .. c
	if self:HasModule(playerClassModule) then
		self:ToggleModule(playerClassModule, true)
		self.activeModule = self:GetModule(playerClassModule)
	end
end

function OmenClassModuleCore:GetThreat()
	return self.activeModule:GetThreat()
end

function OmenClassModuleCore:ClearThreat()
	return self.activeModule:ClearThreat()
end
	
--[[ Module prototype ]]--
function OmenClassModuleCore.modulePrototype:OnEnable()
	self.parser = ParserLib:GetInstance("1.1")
	
	self.buffThreatModifiers = 1
	self.debuffThreatModifiers = 1
	self.equipmentThreatModifiers = 1
	self.equipmentBonusThreatModifiers = 1
	
	self.itemSets = {}
	self.itemSetsWorn = {}
	
	self.myName = UnitName("player")
	
	-- To be filled in by each class module
	self.ClassBuffs = {}
	self.ClassDebuffs = {}
	self.CastHandlers = {}
	self.totalThreatMods = nil
	self.abilityThreatMods = {}
	
	self.generalThreat = 0
	self.targetThreat = 0

	self:ClassInit()
	self:RegisterEvent("SpecialEvents_PlayerBuffGained", function(buffName, buffIndex, applications, texture, rank, index) self:buffGained(buffName, buffIndex, applications, texture, rank, index) end)
	self:RegisterEvent("SpecialEvents_PlayerBuffLost", function (buffName, applications, texture, rank) self:buffLost(buffName, applications, texture, rank) end)
	self:RegisterEvent("SpecialEvents_PlayerDebuffGained", function(debuffName, applications, debuffType, texture, rank, index) self:debuffGained(debuffName, applications, debuffType, texture, rank, index) end)
	self:RegisterEvent("SpecialEvents_PlayerDebuffLost", function(debuffName, applications, debuffType, texture, rank) self:debuffLost(debuffName, applications, debuffType, texture, rank) end)
	self:RegisterEvent("SpecialEvents_EquipmentChanged", function(index, link, oldLink) self:equipChanged(index, link, oldLink) end)
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")

	local events = {
		'CHAT_MSG_SPELL_PERIODIC_PARTY_BUFFS',
		'CHAT_MSG_SPELL_SELF_BUFF',
		'CHAT_MSG_SPELL_SELF_DAMAGE',
		'CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS',
		'CHAT_MSG_SPELL_CREATURE_VS_SELF_BUFF',
		'CHAT_MSG_COMBAT_SELF_HITS',
		'CHAT_MSG_COMBAT_PET_HITS',
		'CHAT_MSG_SPELL_DAMAGESHIELDS_ON_SELF',
		'CHAT_MSG_COMBAT_CREATURE_VS_SELF_HITS',
		'CHAT_MSG_SPELL_PERIODIC_SELF_DAMAGE',
		'CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS',
	}
	for k,v in pairs(events) do
		self.parser:RegisterEvent("OmenClassModule" .. self.className, v, function(evtType,evtInfo) self:parseCombatEvent(evtType,evtInfo) end)
	end
end

function OmenClassModuleCore.modulePrototype:OnDisable() end

function OmenClassModuleCore.modulePrototype:getWornSetPieces(name)
	if self.itemSetsWorn[name] then return self.itemSetsWorn[name] end
	
	local ct = 0
	if self.itemSets[name] then
		for i = 1,12 do
			local itemlink = GetInventoryItemLink("player", i)
			for k,v in pairs(self.itemSets[name]) do
				if string.find(itemlink, v) then
					ct = ct + 1
				end
			end
		end
	end
	self.itemSetsWorn[name] = ct
	return ct
end

function OmenClassModuleCore.modulePrototype:DeathModifier()
	if UnitIsDead("player") then return 0 end
	return 1
end

--[[ Event handling ]]--
function OmenClassModuleCore.modulePrototype:equipChanged(index, link, oldLink)
	for k,v in pairs(self.itemSetsWorn) do self.itemSetsWorn[k] = nil end
	self.totalThreatMods = nil
	
	-- Scan equip for any modifying items
	-- Scan equip for any modifying set bonuses
	-- Scan equip for any modifying enchants (2% threat gloves, -2% threat cloak)
end

function OmenClassModuleCore.modulePrototype:calculateBuffThreatMod()
	self.buffThreatModifiers = 1
	for buffName, index, applications, texture, rank in Aura:BuffIter("player") do
		if Omen.BuffModifiers[buffName] then
			self.buffThreatModifiers = Omen.BuffModifiers[buffName](self, self.buffThreatModifiers)
		elseif self.ClassBuffs[buffName] then
			self.buffThreatModifiers = self.ClassBuffs[buffName](self, self.buffThreatModifiers)
		end
	end
	self.totalThreatMods = nil
end

function OmenClassModuleCore.modulePrototype:calculateDebuffThreatMod()
	self.debuffThreatModifiers = 1
	for buffName, index, applications, texture, rank in Aura:DebuffIter("player") do
		if OmenDebuffModifiers[buffName] then
			self.debuffThreatModifiers = OmenDebuffModifiers[buffName](self, self.debuffThreatModifiers)
		elseif self.ClassDebuffs[buffName] then
			self.debuffThreatModifiers = self.ClassDebuffs[buffName](self, self.debuffThreatModifiers)
		end
	end
	self.totalThreatMods = nil
end

-- Update buff/debuff modifiers when auras change
function OmenClassModuleCore.modulePrototype:buffGained(buffName, buffIndex, applications, texture, rank, index) 
	if OmenBuffModifiers[buffName] or self.ClassBuffs[buffName] then self:calculateBuffThreatMod() end
end

function OmenClassModuleCore.modulePrototype:buffLost(buffName, applications, texture, rank) 
	if OmenBuffModifiers[buffName] or self.ClassBuffs[buffName] then self:calculateBuffThreatMod() end 
end

function OmenClassModuleCore.modulePrototype:debuffGained(debuffName, applications, debuffType, texture, rank, index)
	if OmenDebuffModifiers[buffName] or self.ClassDebuffs[buffName] then self:calculateDebuffThreatMod() end
end

function OmenClassModuleCore.modulePrototype:debuffLost(debuffName, applications, debuffType, texture, rank)
	if OmenDebuffModifiers[buffName] or self.ClassDebuffs[buffName] then self:calculateDebuffThreatMod() end
end

function OmenClassModuleCore.modulePrototype:ClassInit()
	-- By default, we assume threat to be unmodified
	self.passiveThreatModifiers = 1
end

function OmenClassModuleCore.modulePrototype:parseCombatEvent(t, info)
	if info.type == "hit" then self:parseDamage(t, info)
	elseif info.type == "heal" then self:parseHeal(t, info)
	elseif info.type == "gain" then self:parseGain(t, info)
	end
end

-- Basic damage and healing parsing
function OmenClassModuleCore.modulePrototype:parseDamage(t, info)
	if info.source ~= ParserLib_SELF then return end
	
	local threat = info.amount
	if(self.abilityThreatMods[info.skill]) then threat = self.abilityThreatMods[info.skill](self, threat) end
	threat = threat * self:threatMods()
	self:AddTargetThreat(threat, info.victim)
end

function OmenClassModuleCore.modulePrototype:parseHeal(t, info)
	if info.source ~= ParserLib_SELF then return end
	
	local threat = info.amount
	if(self.abilityThreatMods[info.skill]) then threat = self.abilityThreatMods[info.skill](self, threat) end
	threat = threat * Omen.ThreatConstants.healing * self:threatMods()
	self:AddThreat(threat)
end

function OmenClassModuleCore.modulePrototype:parseGain(t, info)
	if info.victim ~= ParserLib_SELF then return end
	
	local maxgain = UnitManaMax("player") - UnitMana("player")
	local amount = math.min(maxgain, info.amount)
	
	if info.attribute == "Mana" and not self:ExemptGain(info.skill) then
		self:AddThreat(amount * Omen.ThreatConstants.managain * self:threatMods())
	elseif info.attribute == "Rage" and not self:ExemptGain(info.skill) then
		self:AddThreat(amount * Omen.ThreatConstants.ragegain * self:threatMods())
	elseif info.attribute == "Energy" and not self:ExemptGain(info.skill) then
		self:AddThreat(amount * Omen.ThreatConstants.energygain * self:threatMods())
	end
end


function OmenClassModuleCore.modulePrototype:UNIT_SPELLCAST_SUCCEEDED()
	if arg1 ~= "player" then return end
	if self.CastHandlers[arg2] then self.CastHandlers[arg2](self, arg3) end
end

function OmenClassModuleCore.modulePrototype:PLAYER_REGEN_ENABLED()
	self:ClearThreat()
end

--[[ End parsing ]]--

function OmenClassModuleCore.modulePrototype:threatMods()
	if self.totalThreatMods ~= nil then
		return self.totalThreatMods
	end
	self.totalThreatMods = self.buffThreatModifiers * self.passiveThreatModifiers * self.debuffThreatModifiers * self.equipmentThreatModifiers * self.equipmentBonusThreatModifiers
	return self.totalThreatMods
end

function OmenClassModuleCore.modulePrototype:GetThreat()
	return (self.targetThreat * self:DeathModifier()), (self.generalThreat * self:DeathModifier()), 0
end

function OmenClassModuleCore.modulePrototype:ClearThreat()
	self.generalThreat = 0
	self.targetThreat = 0
end

function OmenClassModuleCore.modulePrototype:AddTargetThreat(threat)
	self.targetThreat = self.targetThreat + threat
	Omen:Debug(sprintf("Added target threat. Total is now %s", self.targetThreat))
end

-- This will end up being regen + heals, pretty much
function OmenClassModuleCore.modulePrototype:AddThreat(threat)
	self.generalThreat = self.generalThreat + threat
	Omen:Debug(sprintf("Added general threat. Total is now %s", self.generalThreat))
end

function OmenClassModuleCore.modulePrototype:ExemptGain(skill)
	return false
end